//
// fs_classic00 フィールドサーバシールドスケッチ
// fs-classic用Arduino Program by T.Fukatsu
// 1時間に10分(**:54～**:03)起動。Twitter,Webサーバ,ファンOn
// D8pinがHのときは常時稼働モード。LED2点灯
// 電源投入時にLED1がonでNTPにアクセスできたら消灯
// http://***/?reset=onでリセット, relay2=on/offでカメラ電源
// relay1=offでファン電源off, relay0=offで無線LAN off
//
// 2013/7/1 ver.1.1
// 2013/8/3 ver.1.12
// 2013/8/4 ver.1.13 なぜか数回ループするとスリープ復帰後ハング？初回のみNTP
// 2013/8/6 ver.1.14 Webページ header使うとエージェント不可,旧版戻す
// 2013/9/4 ver.1.15 ソフトリセット追加
// 2013/10/18 ver.1.16 header分割、テストモード時割り込み無
// 2013/11/1 ver.1.17 時刻部分修正、Etherのreset修正
// 2014/3/4 ver.1.172 準常時稼働モード,NTP修正
// 2014/5/18 ver.1.173 パラメータ変更,Fan_Wait追加
//
#include <SPI.h>
#include <Ethernet.h>
#include <MsTimer2.h>
#include <Twitter.h>
#include <Time.h>
#include <EthernetUdp.h>

// Web,Twitter用雛型文章
PROGMEM prog_char version_name[] = "[FS-c00 v1.173]";
PROGMEM prog_char web_header[] = "HTTP/1.0 200 OK\nContent-Type: text/html\nConnection: close\n";
PROGMEM prog_char web_top1[]    = "<html><head><title>Field Server Classic Web Server</title></";
PROGMEM prog_char web_top2[]    = "head><body><h3>Field Server Classic ver.1.173</h3><hr/><form>";
PROGMEM prog_char web_top3[]    = "<table border>";
PROGMEM prog_char web_end[] = "</table></form><HR/>2013.9.4 by T.FUKATSU(NARO)</body></html>";

// 間欠動作時間に関係するパラメータ
const unsigned long device_Wait = 1000UL*2UL; //(ms) 動作安定のためのウェイト
const unsigned long wireless_Wait = 1000UL*62UL; //(ms) 無線/3G起動待ちのウェイト
const unsigned long fan_Wait = 1000UL*60UL*0UL; //(ms) センサ安定のウェイト
const unsigned int wake_up_Time = 60*54; //(sec) 00分から数えてこの時間に起動
const unsigned int wake_Period = 60*10; //(sec) 起動している時間

// Arduinoのpin設定
const byte DC5V_pin = 2;       //Wireless and EtherShield DC/DC converter(CN10) Relay
const byte DC12V_pin1 = 3;     //FAN control DC/DC converter(CN7) Relay
const byte DC12V_pin2 = 4;     //camera control DC/DC converter(CN8) Relay
const byte reset_pin = 5;      //for Ethernet Shield reset(起動時にhighにする必要有)
const byte LED_pin1 = 6;       //output LED, 起動時onでNTPアクセス成功時off
const byte LED_pin2 = 7;       //output LED, これがonの時は常時稼働モード
const byte MODE_pin = 8;       //H:常時稼働(テスト用) <--> L:インターバルモード

// インターネット接続関係 PRGMEMで書くと動作せず...
EthernetServer server(80);
byte mac[] = { 0x90, 0xA2, 0xDA, 0x0D, 0xA4, 0xA4 };
byte  ip[] = { 192, 168, 10, 200 }; //テスト用IP
// byte dns_ip[] = { 192, 168, 100, 1 }; //dnsという名前だとエラー
// byte gateway[] = { 192, 168, 100, 1 }; //最後が*.1でない場合は設定必要
// byte subnet[] = { 255, 255, 255, 0 };

// NTPサーバアクセス関係
const unsigned int localPort = 8888;
//IPAddress timeServer(202, 224, 32, 4); //ntp.asahi-net.or.jp 現在停止中?
IPAddress timeServer(133, 243, 238, 164); //ntp.nict.jp
//IPAddress timeServer(150, 26, 2, 66); //stm1.cc.affrc.go.jp(maffin内部のみ)
const byte NTP_PACKET_SIZE = 48;
byte packetBuffer[NTP_PACKET_SIZE];
EthernetUDP Udp;

//Twitterのトークン
int twitte_status = 0;
Twitter twitter("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"); // fs-classic00

//データ取得時の変数
unsigned int raw[6]; // A/Dで得られた生電圧を保存(0-1023)
char str[64]; // コメント表示,Twitterなどテンポラリ文字列を格納
char msg[141]; // Twitteする本文を格納,最大140文字
String arg_data=""; //Webサーバアクセス時の引数を格納

// 制御用フラグ
boolean flag = true; //これがfalseになるとloop()は終了し、再度loop()を実行
boolean mode_flag = false; //interval動作はfalse, trueにすると常時稼働
boolean twitter_flag = true; //twitteしないときはfalse(所内ネットからはtwitte不可)
boolean reset_flag = false; //trueのときsoftReset. Web上でreset=onとしたとき用
boolean relay2_flag = false; //Web表示用
boolean relay1_flag = false; //Web表示用

// NTPおよび内部時刻保存用
int NTP_time = 0; //(s) from 00 min.からの経過秒数 (求められなかった時は-1)
const int INITIAL_OFFSET = 3000; //(ms):NTPアクセス後、指定時間になるまでのウェイトオフセット
const int DEFAULT_OFFSET = 500; //(ms)割り込み毎のウェイトオフセット
const byte DEFAULT_RESET_HOUR = 5; //(0-23) この時間起動後のスリープ時(wake_up_Time+wake_Period)にリセット
byte offline_hour = 5; //(0-23) NTPにアクセス出来なかったときでも時間が分かるようカウント


//
// セットアップ
//
void setup(){ 

  // initial  
  Serial.begin(9600); 
  strcpy_P(str,(char*)version_name);
  Serial.println(str);

  // pin setup
  pinMode (DC5V_pin, OUTPUT);
  pinMode (DC12V_pin1, OUTPUT);
  pinMode (DC12V_pin2, OUTPUT);
  pinMode (reset_pin, OUTPUT);
  pinMode (LED_pin1, OUTPUT);
  pinMode (LED_pin2, OUTPUT);
  pinMode (MODE_pin, INPUT);

  // power on
  digitalWrite(reset_pin, LOW); //最初はLOWに設定
  digitalWrite(LED_pin1, HIGH); //チェック用LEDを点灯
  digitalWrite(DC5V_pin, HIGH); //Etherシールド&無線LANの電源ON
  digitalWrite(DC12V_pin1, HIGH); //ファンを稼働
  relay1_flag = true;
  digitalWrite(DC12V_pin2, LOW); //カメラは普段は電源OFF
  relay2_flag = false;
  if( digitalRead(MODE_pin) == HIGH ){
    mode_flag = true; //MODE_pin(SW3)が短絡してHIGHの時は常時稼働
    digitalWrite(LED_pin2, HIGH); //チェック用LEDを点灯
  } else {
    mode_flag = false;
    digitalWrite(LED_pin2, LOW); //ソフトリセット時前の状態が残るため
  }
  delay(device_Wait); //不要かも？
  digitalWrite(reset_pin, HIGH); //Etherシールドをリセットしないとうまく動作しない
  analogReference(INTERNAL); //アナログ入力は0-1Vに切り替え

  // wireless
  Serial.println(F("waiting..."));
  delay(wireless_Wait);
  Ethernet.begin(mac, ip);
  // Ethernet.begin(mac, ip, dns_ip, gateway, subnet); //フルに記述する場合
  Udp.begin(localPort);
  delay(device_Wait);
  sendNTPpacket(timeServer); //AccessNTP内でも実行しているから不要かも？
  delay(device_Wait); //NTPサーバにうまくアクセスするため必要
 
  //NTP access
  AccessNTP();
  Udp.stop();
  int initial_sleep = 0; //(sec) 初期設定後、指定時間になるまでウェイトする時間
  if( NTP_time != -1 ){
    digitalWrite(LED_pin1, LOW); //チェック用LEDを消灯
    //現在時刻から指定起動時間までどれくらいか計算
    if( wake_up_Time > (NTP_time+INITIAL_OFFSET/1000) ){
      offline_hour = hour()-1; //loop()で+1加算されるのでここで-1しておく
      initial_sleep = wake_up_Time - NTP_time;
    } else {
      offline_hour = hour();
      initial_sleep = 3600 + wake_up_Time - NTP_time;
    }     
  } else {
    //NTP失敗時はタイマーのソフトリセットが掛ったと想定して設定
    initial_sleep = 3600 - wake_Period;
    offline_hour = DEFAULT_RESET_HOUR;
    if( (wake_up_Time + wake_Period) >= 3600 ) offline_hour++;
    if( offline_hour >= 24 ) offline_hour = 0;
  } 
  
  //Do something
  //きちんとNTPで時刻補正できたかをTwitte
  strcpy_P(msg, (char*)version_name);
  sprintf(str, " %d//%d (s)", NTP_time, initial_sleep);
  strcat(msg, str);
  sprintf(str, " --> wake up %d (s) from 00", wake_up_Time);
  strcat(msg, str);
  Serial.println(msg);
  if(twitter_flag) ArduinoTwitte(msg);
  
  //一端全ての電源を落として正しい起動時間になるまでスリープ(通常モード)
  if( !mode_flag ){
    digitalWrite(DC12V_pin2, LOW); 
    relay2_flag = false; 
    digitalWrite(DC12V_pin1, LOW);
    relay1_flag = false;
    digitalWrite(DC5V_pin, LOW);
    digitalWrite(reset_pin, LOW);    
    delay(initial_sleep*1000UL-INITIAL_OFFSET); 
    // 指定した起動時間経過毎に割り込みをかけてスリープする 
    MsTimer2::set(wake_Period*1000UL, Interrupt_function);
    MsTimer2::start();
  } 
}

//
// 割り込みルーチン, 基本的にはスリープ
//
void  Interrupt_function(){
  interrupts(); //delayを使用するためには必要
  MsTimer2::stop(); //割り込みが終わるまで次の割り込みまでのカウントを停止
  if( offline_hour == DEFAULT_RESET_HOUR ) softReset(); //1日1回12:50起動時(->13:00)にリセット
  digitalWrite(DC12V_pin2, LOW);
  relay2_flag = false;
  digitalWrite(DC12V_pin1, LOW);
  relay1_flag = false;
  digitalWrite(DC5V_pin, LOW);
  digitalWrite(reset_pin, LOW);
  Udp.stop(); //念のため(不要?)
  //sleep_Offsetは毎時NTPにアクセスしなくなったので削除
  Serial.print(F("sleep[s]:"));
  Serial.println(3600-wake_Period);
  delay( (3600-wake_Period)*1000UL - DEFAULT_OFFSET ); 
  Serial.println(F("back to loop"));
  flag = false; // loop内の処理を一端終了させる
  MsTimer2::start();
}

//
// メインループ
//
void loop(){
  // 初期設定
  flag = true;
  reset_flag = false;
  sprintf(msg,""); // 念のため初期化
  sprintf(str,""); // 念のため初期化
  
  if( !mode_flag ){
    // power on
    digitalWrite(DC12V_pin1, HIGH); //ファンを稼働
    delay(fan_Wait);
    digitalWrite(DC5V_pin, HIGH); //Etherシールド&無線LANの電源ON    
    relay1_flag = true;
    delay(device_Wait); //不要かも？
    digitalWrite(reset_pin, HIGH); //Etherシールドをリセットしないとうまく動作しない
    // wireless
    Serial.println(F("waiting..."));
    delay(wireless_Wait);
    Ethernet.begin(mac, ip);
    // Ethernet.begin(mac, ip, dns_ip, gateway, subnet); //フルに記述する場合
  }

  // ローカル時刻更新
  offline_hour++; //前回に比べて1時間経過
  if( offline_hour >=24 ) offline_hour = 0;

  //Do something
  //Twitteに書き込む内容を作成
  strcpy_P(msg, (char*)version_name);
  sprintf(str, "[%lu] %d(h):%d(s) ", millis()/1000, offline_hour, wake_up_Time);
  strcat(msg,str);
  ReadSensor();
  sprintf(str ,"ad0:%d, ad1:%d, ad2:%d, ad3:%d, ad4:%d, ad5:%d",
                raw[0], raw[1], raw[2], raw[3], raw[4], raw[5]); 
  strcat(msg, str);
  Serial.println(msg);
  if(twitter_flag) ArduinoTwitte(msg);


  // 11:50になったらカメラの電源をOn(1日1回稼働)
  if( offline_hour == 7 || offline_hour == 11 || offline_hour == 13 ){
    digitalWrite(DC12V_pin2,HIGH);
    relay2_flag = true;
  }
/* 
    // カメラ電源on
    digitalWrite(DC12V_pin2,HIGH);
    relay2_flag = true;
*/    

  // Web server start
   server.begin();
   while(flag){
      EthernetClient client = server.available();
      arg_data = ""; //Webサーバアクセス時の引数を初期化
      boolean arg_flag = true; //HTTPのヘッダを読み込み終えるまでtrue
      if (client && flag) {
        // an http request ends with a blank line
         boolean currentLineIsBlank = true;
         while (client.connected() && flag) {
            delay(1); //timing delay else webserver hangs.
            if (client.available() && flag) {
              char c = client.read();
              if( arg_flag ) arg_data.concat(c);
              if (c == '\n' && currentLineIsBlank) {
                //引数を確認して動作
                //カメラの電源制御
                if(arg_data.indexOf("relay2=off") != -1 ){
                  digitalWrite(DC12V_pin2,LOW);
                  relay2_flag = false;
                }
                if(arg_data.indexOf("relay2=on") != -1 ){
                  digitalWrite(DC12V_pin2,HIGH);
                  relay2_flag = true;
                }
                //ファンの電源制御 (Offのみ)                 
                if(arg_data.indexOf("relay1=off") != -1 ){
                  digitalWrite(DC12V_pin1,LOW);
                  relay1_flag = false;
                }
                //無線の電源制御 (Offのみ)
                if(arg_data.indexOf("relay0=off") != -1 ){
                  digitalWrite(DC5V_pin,LOW);
                }
                //機器のソフトウェアリセット
                if(arg_data.indexOf("reset=on") != -1 ){
                  reset_flag = true;
                }
   
                // Webレスポンス
                strcpy_P(str, (char*)web_header);
                client.println(str);                
                strcpy_P(str, (char*)web_top1);
                client.print(str);
                strcpy_P(str, (char*)web_top2);
                client.print(str);
                strcpy_P(str, (char*)web_top3);
                client.println(str);
                client.print(F("<tr><td>"));                
                strcpy_P(msg, (char*)version_name);  
                unsigned long sec = millis()/1000;                
                sprintf(str, "[%lu sec]", sec);
                strcat(msg,str);
                client.print(msg);
                client.println(F(" </td></tr>"));               
                ReadSensor();             
                for (int analogChannel = 0; analogChannel < 6; analogChannel++) {
                  client.print(F("<tr><td>ad"));
                  client.print(analogChannel);
                  client.print(F(":"));
                  client.print(raw[analogChannel]);
                  client.println(F("<br /></td></tr>"));
                }
                client.print(F("<tr><td>&nbsp;</td></tr>"));
                //カメラ(relay2)操作ボタン
                client.print(F("<tr><td>cam: <input type=\"submit\" name=\"relay2\" value=\"on\">"));
                if( relay2_flag ){
                  client.print(F("on"));
                } else {
                  client.print(F("off"));
                }
                client.print(F("<input type=\"submit\" name=\"relay2\" value=\"off\"></td></tr>"));
                //リロードボタン
                client.print(F("<tr><td><p align=\"center\"><input type=\"submit\" name=\"\" value=\"reload\"></p></td></tr>"));
                // Twitter状態
                client.print(F("<tr><td>twitter:"));
                if(twitter_flag){
                  client.print(twitte_status);
                } else {
                  client.print(F("---"));
                }
                client.println(F("</td></tr>"));
                // 残り                
                strcpy_P(str, (char*)web_end);
                client.println(str);
                Serial.println(F("http OK."));
                break;
              }
              if (c == '\n') {
                // you're starting a new line
                arg_flag = false;
                currentLineIsBlank = true;
              } 
              else if (c != '\r') {
                // you've gotten a character on the current line
                currentLineIsBlank = false;
              }
           }
         }
         // give the web browser time to receive the data
         delay(1);
         // close the connection:
         client.stop();
         if( reset_flag ) softReset();       
      }
   }
   Serial.println(F("loop finish")); 
}

//
// NTPリクエストを送信
//
unsigned long sendNTPpacket(IPAddress& address)
{
   // set all bytes in the buffer to 0
   memset(packetBuffer, 0, NTP_PACKET_SIZE); 
   // Initialize values needed to form NTP request
   // (see URL above for details on the packets)
   packetBuffer[0] = 0b11100011;   // LI, Version, Mode
   packetBuffer[1] = 0;     // Stratum, or type of clock
   packetBuffer[2] = 6;     // Polling Interval
   packetBuffer[3] = 0xEC;  // Peer Clock Precision
   // 8 bytes of zero for Root Delay & Root Dispersion
   packetBuffer[12]  = 49; 
   packetBuffer[13]  = 0x4E;
   packetBuffer[14]  = 49;
   packetBuffer[15]  = 52;
   // all NTP fields have been given values, now
   // you can send a packet requesting a timestamp:         
   Udp.beginPacket(address, 123); //NTP requests are to port 123
   Udp.write(packetBuffer,NTP_PACKET_SIZE);
   Udp.endPacket(); 
}

//
// NTPで時刻を取得
//
void AccessNTP(){
  Serial.println(F("send NTP packet"));
  sendNTPpacket(timeServer);
  delay(1000); //念のため(信州大のコードを参照)
  if ( Udp.parsePacket() ) {
     Udp.read(packetBuffer, NTP_PACKET_SIZE);
     // 時刻情報はパケットの40バイト目からはじまる4バイトのデータ
     unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
     unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
     // NTPタイムスタンプは64ビットの符号無し固定小数点数
     //（整数部32ビット、小数部32ビット）
     // 1900年1月1日0時との相対的な差を秒単位で表している
     // 小数部は切り捨てて、秒を求めている
     unsigned long secsSince1900 = highWord << 16 | lowWord;
     // NTPタイムスタンプをUNIXタイムに変換する
     // UNITタイムは1970年1月1日0時からはじまる
     // 1900年から1970年の70年を秒で表すと2208988800秒になる
     const unsigned long seventyYears = 2208988800UL;
     // NTPタイムスタンプから70年分の秒を引くとUNIXタイムが得られる
     unsigned long epoch = secsSince1900 - seventyYears;
     // Timeライブラリに時間を設定(UNIXタイム)
     // 日本標準時にあわせるために＋9時間しておく
     setTime(epoch + (9 * 60 * 60));
//     sprintf(str, "JST is %d / %d / %d", year(), month(), day());
//     Serial.print(str);
     sprintf(str, " %d : %d : %d", hour(), minute(), second());
     Serial.println(str);
     NTP_time = second()+minute()*60; //00分からの秒数を算出
  } else {
    Serial.println(F("NTP failed."));
    NTP_time = -1;
  } 
}

//
// A/D接続のセンサ値を読み込む(8回取りこんで平均する)
//
void ReadSensor(){
   for(byte analogChannel=0; analogChannel<6; analogChannel++) {
      analogRead(analogChannel); // dummy,チャンネル間の電圧吸い込み誤差を低減
      raw[analogChannel] = 0;
      for( byte i=0; i<8; i++){
         raw[analogChannel] += analogRead(analogChannel);
      }
      raw[analogChannel] = (raw[analogChannel]+4) >> 3; //切捨なので8/2を足してから8で割る
   }
}

//
// msgの内容をtwitteする
//
void ArduinoTwitte( char msg[141] )
{
  Serial.println(F("connecting ..."));
  if (twitter.post(msg)) {
    twitte_status = twitter.wait();
    if (twitte_status == 200) {
        Serial.println(F("twitte OK"));
    } else {
      Serial.print(F("failed : code "));
      Serial.println(twitte_status);
    }
  } else {
    twitte_status = -1;
    Serial.println(F("connection failed."));
  }
  Serial.println(F("twitte end"));
}

/*
//
// デバッグ用メモリチェック
//
void check_mem()
{
  uint8_t *stackptr, *heapptr;
  stackptr = (uint8_t *)malloc(4);    //use stackptr temporarily
  heapptr = stackptr;                 //save value of heap pointer
  free(stackptr);                     //set stackptr to 0
  stackptr = (uint8_t *)(SP);         //save value of stack pointer
  
  Serial.print(F("Stack ptr :"));
  Serial.print((long)stackptr);
  Serial.print(F("<--> Heap ptr :"));
  Serial.println((long)heapptr);     //stackptr - heapptrが残りメモリ量
}
*/

//
// ソフトウェアリセット 
//
void softReset()
{
  Serial.println("Reboot!");
  MsTimer2::stop(); //MsTimer2::start()のままだとハングする...
  delay(1000);
  asm volatile ("  jmp 0");  
}
